import uuid
from datetime import datetime
from typing import Any, Dict, List

from fastapi.exceptions import HTTPException
from starlette import status
from tinydb import Query, TinyDB

NOT_FOUND = HTTPException(
    status_code=status.HTTP_404_NOT_FOUND, detail="Entity not found",
)

query = Query()


def create(db: TinyDB, entity: Dict[str, Any]) -> Dict[str, Any]:
    entity["id"] = str(uuid.uuid4())
    entity["createdTime"] = datetime.now().timestamp()
    entity["updatedTime"] = entity["createdTime"]

    db.insert(entity)
    return entity


def read(db: TinyDB, id: str) -> Dict[str, Any]:
    matching_entities = db.search(query.id == id)

    if not matching_entities:
        raise NOT_FOUND

    return matching_entities[0]


def read_all(db: TinyDB) -> List[Dict[str, Any]]:
    return db.all()


def update(db: TinyDB, id: str, entity: Dict[str, Any]) -> Dict[str, Any]:
    matching_entity = read(db, id)

    entity["id"] = id
    entity["createdTime"] = matching_entity["createdTime"]
    entity["updatedTime"] = datetime.now().timestamp()

    db.update(entity, query.id == id)
    return entity


def delete(db: TinyDB, id: str) -> None:
    read(db, id)
    db.remove(query.id == id)


def delete_all(db: TinyDB) -> None:
    db.drop_tables()



from typing import Any, Dict, Generic, List, Optional, TypeVar

from pydantic import BaseModel
from pydantic.generics import GenericModel

ReturnData = TypeVar("ReturnData")


class JsonApiSource(BaseModel):
    pointer: Optional[str]
    parameter: Optional[str]
    meta: Optional[Dict[Any, Any]]


class JsonApiErrorObject(BaseModel):
    id: Optional[str]
    links: Optional[Dict[Any, Any]]
    about: Optional[str]
    status: Optional[int]
    title: Optional[str]
    detail: Optional[str]
    source: Optional[JsonApiSource]
    meta: Optional[Dict[Any, Any]]


class JsonApiSuccessResponse(GenericModel, Generic[ReturnData]):
    data: ReturnData
    jsonapi: Optional[Dict[Any, Any]] = None
    meta: Optional[Dict[Any, Any]] = None


class JsonApiErrorResponse(GenericModel, Generic[ReturnData]):
    jsonapi: Optional[Dict[Any, Any]]
    meta: Optional[Dict[Any, Any]]
    errors: Optional[List[JsonApiErrorObject]]


class JsonApiRequestEntity(GenericModel, Generic[ReturnData]):
    data: ReturnData



from typing import Any, Callable, Generic, List, Type, TypeVar

from fastapi import APIRouter
from pydantic import BaseModel
from starlette import status

from ._json_api import JsonApiSuccessResponse

T = TypeVar("T", bound=BaseModel)  # noqa: WPS111


class CRUDGenerator(Generic[T], APIRouter):  # noqa: WPS214
    schema: Type[T]
    create_schema: Type[T]
    _base_path: str = "/"

    def __init__(self, schema: Type[T], create_schema: Type[T], **kwargs: Any) -> None:

        self.schema = schema
        self.create_schema = create_schema

        schema_name = self.schema.__name__

        prefix = "{path}{schema}s".format(
            path=self._base_path, schema=self.schema.__name__.lower().strip("/"),
        )
        tags = [prefix.strip("/")]

        super().__init__(prefix=prefix, tags=tags, **kwargs)  # noqa: WPS613

        self._add_api_route(
            "",
            self._create(),
            methods=["POST"],
            name="{0}s:create_{0}".format(schema_name),
            summary="Creates a new {0}".format(schema_name),
            status_code=status.HTTP_201_CREATED,
            response_model=JsonApiSuccessResponse[self.create_schema],
        )

        self._add_api_route(
            "/{entity_id}",
            self._read(),
            methods=["GET"],
            name="{0}s:get_{0}".format(schema_name),
            summary="Gets an existing {0} by id".format(schema_name),
            status_code=status.HTTP_200_OK,
            response_model=JsonApiSuccessResponse[self.schema],
        )

        self._add_api_route(
            "",
            self._read_all(),
            methods=["GET"],
            name="{0}s:get_all".format(schema_name),
            summary="Gets all existing {0} entities in the system".format(schema_name),
            status_code=status.HTTP_200_OK,
            response_model=JsonApiSuccessResponse[List[self.schema]],
            response_model_exclude_unset=True,
        )

        self._add_api_route(
            "/{entity_id}",
            self._update(),
            methods=["PATCH"],
            name="{0}s:update_{0}".format(schema_name),
            summary="Updates an existing {0}".format(schema_name),
            status_code=status.HTTP_200_OK,
            response_model=JsonApiSuccessResponse[self.create_schema],
        )

        self._add_api_route(
            "/{entity_id}",
            self._delete(),
            methods=["DELETE"],
            name="{0}s:delete_{0}".format(schema_name),
            summary="Deletes an existing {0}".format(schema_name),
            status_code=status.HTTP_204_NO_CONTENT,
        )

        self._add_api_route(
            "",
            self._delete_all(),
            methods=["DELETE"],
            name="{0}s:delete_all".format(schema_name),
            summary="Deletes all existing {0} entities in the system".format(
                schema_name,
            ),
            status_code=status.HTTP_204_NO_CONTENT,
        )

    def _add_api_route(
        self, path: str, endpoint: Callable[..., Any], **kwargs: Any,
    ) -> None:
        super().add_api_route(path, endpoint, **kwargs)  # noqa: WPS613

    # These should be overriden by whatever is extending this class.
    def _create(self, *args: Any, **kwargs: Any) -> Callable[..., Any]:
        raise NotImplementedError

    def _read(self, *args: Any, **kwargs: Any) -> Callable[..., Any]:
        raise NotImplementedError

    def _read_all(self, *args: Any, **kwargs: Any) -> Callable[..., Any]:
        raise NotImplementedError

    def _update(self, *args: Any, **kwargs: Any) -> Callable[..., Any]:
        raise NotImplementedError

    def _delete(self, *args: Any, **kwargs: Any) -> Callable[..., Any]:
        raise NotImplementedError

    def _delete_all(self, *args: Any, **kwargs: Any) -> Callable[..., Any]:
        raise NotImplementedError



from typing import Any, Callable, Coroutine, Dict, List, Type

from fastapi import Depends
from fastapi.encoders import jsonable_encoder
from loguru import logger
from pydantic import BaseModel
from tinydb import TinyDB

from ._base import CRUDGenerator
from . import _tinydb_crud as crud
from ._json_api import JsonApiRequestEntity

CALLABLE = Callable[..., Coroutine[Any, Any, Dict[str, Any]]]
CALLABLE_LIST = Callable[..., Coroutine[Any, Any, List[Dict[str, Any]]]]  # noqa: WPS221


class TinydbCRUDRouter(CRUDGenerator[BaseModel]):
    def __init__(
        self,
        db: "TinyDB",
        schema: Type[BaseModel],
        create_schema: Type[BaseModel],
        **kwargs: Any,
    ) -> None:

        self.schema_name = schema.__name__
        self.db_func = db

        super().__init__(
            schema=schema, create_schema=create_schema, **kwargs,
        )

    def _create(self, *args: Any, **kwargs: Any) -> CALLABLE:
        async def route(
            entity: JsonApiRequestEntity[self.create_schema],
            db: TinyDB = Depends(self.db_func),
        ) -> Dict[str, Any]:
            entity = crud.create(db, jsonable_encoder(entity.data))
            logger.info("Created {0} {1}".format(self.schema_name, entity["id"]))
            return {"data": entity}

        return route

    def _read(self, *args: Any, **kwargs: Any) -> CALLABLE:
        async def route(
            entity_id: str, db: TinyDB = Depends(self.db_func),
        ) -> Dict[str, Any]:
            entity = crud.read(db, entity_id)
            logger.info("Read {0} {1}".format(self.schema_name, entity_id))
            return {"data": entity}

        return route

    def _read_all(self, *args: Any, **kwargs: Any) -> CALLABLE_LIST:
        async def route(
            db: TinyDB = Depends(self.db_func),
        ) -> Dict[str, List[Dict[str, Any]]]:
            entities = crud.read_all(db)
            logger.info("Read all {0}s".format(self.schema_name))
            return {"data": entities}

        return route

    def _update(self, *args: Any, **kwargs: Any) -> CALLABLE:
        async def route(
            entity_id: str,
            entity: JsonApiRequestEntity[self.create_schema],
            db: TinyDB = Depends(self.db_func),
        ) -> Dict[str, Any]:
            entity = crud.update(db, entity_id, jsonable_encoder(entity.data))
            logger.info("Updated {0} {1}".format(self.schema_name, entity_id))
            return {"data": entity}

        return route

    def _delete(self, *args: Any, **kwargs: Any) -> CALLABLE:
        async def route(entity_id: str, db: TinyDB = Depends(self.db_func)) -> None:
            crud.delete(db, entity_id)
            logger.info("Deleted {0} {1}".format(self.schema_name, entity_id))

        return route

    def _delete_all(self, *args: Any, **kwargs: Any) -> CALLABLE_LIST:
        async def route(db: TinyDB = Depends(self.db_func)) -> None:
            crud.delete_all(db)
            logger.info("Deleted all {0}s".format(self.schema_name))

        return route



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['crudrouter']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.62.0,<0.63.0',
 'loguru>=0.5.3,<0.6.0',
 'pydantic>=1.8.2,<2.0.0',
 'tinydb>=4.5.1,<5.0.0',
 'uuid>=1.30,<2.0',
 'uvicorn>=0.11.8,<0.12.0']

setup_kwargs = {
    'name': 'crudrouter',
    'version': '0.1.2',
    'description': '',
    'long_description': None,
    'author': 'Benjamin Harruff',
    'author_email': 'harruff.benjamin@solute.us',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://code.shipyard.blackpearl.us/bma/sscp/crudrouter',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
